"""
function: automatically downloading model weights for the MRISegmentator package
author = "Abhinav Suri, Yan Zhuang"
license = "Please see the license file"
email = "yan.zhuang2@nih.gov or rsummers@mail.cc.nih.gov"
Date: 06/27/2024
"""

import requests
from tqdm import tqdm
import os
from pathlib import Path
import zipfile

def get_weights_dir():
    """
    Creat directory to store model weights
    """
    env_var = 'MRISEGMENTATOR_DIR'
    if os.environ.get('MRISEGMENTATOR_DIR') is not None:
        dir = os.environ.get('MRISEGMENTATOR_DIR')
    else:
        dir = str(Path.home()) + '/.mrisegmentator_weights'
    try:
        if not os.path.isdir(dir):
            print("making totalsegmentator directory at:", dir)
            os.makedirs(dir)
        return dir
    except Exception as e:
        raise FileNotFoundError(f"cannot create directory to store weights at {dir}: ERROR: {e}")

def download_weights(redownload=False):
    """
    download the weights automatically
    """
    url = 'https://nihcc.app.box.com/index.php?rm=box_download_shared_file&shared_name=q6vl3015hteoufz7jll63u3hdqk79li7&file_id=f_1544045874167'
    
    filepath = get_weights_dir() + '/MRISegmentator.zip'
    final_path = get_weights_dir() + '/nnUNetTrainer__nnUNetPlans__3d_fullres_New'
    if os.path.exists(final_path) and redownload == False:
        print("Weights are already downloaded at", final_path)
        return final_path
    if not os.path.exists(final_path) or redownload == True:
        print("saving weights to", get_weights_dir())
        # from https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests 
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get('content-length', 0))
        block_size = 1024
        with tqdm(total=total_size, unit="B", unit_scale=True) as progress_bar:
            with open(filepath, "wb") as file:
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)
        if total_size != 0 and progress_bar.n != total_size:
            raise RuntimeError('Could not download weights')
    if os.path.exists(filepath):
        print("unzipping model weights")
        with zipfile.ZipFile(filepath, 'r') as zip_ref:
            zip_ref.extractall(get_weights_dir())
            if os.path.exists(final_path):
                print("successfully unzipped weights and saved to", final_path, ". Cleaning up...")
                os.remove(filepath)
                return final_path
    else:
        raise FileNotFoundError(f"for some reason the MRISegmentator.zip file was not found in your weights directory at: {get_weights_dir()}")

def force_redownload():
    download_weights(True) 

if __name__ == "__main__":
    download_weights()



"""
function: specifying arguments for the MRISegmentator package
author = "Abhinav Suri, Yan Zhuang"
license = "Please see the license file"
email = "yan.zhuang2@nih.gov or rsummers@mail.cc.nih.gov"
Date: 06/27/2024
"""

import argparse

def main():
    parser = argparse.ArgumentParser(description="Segment 68 anatomical structures on T1 weighted MRI. Model by Yan Zhuang. Package by Abhinav Suri. If you use this tool cite https://arxiv.org/abs/2405.05944")

    parser.add_argument("-i", metavar="filepath", dest="input",
                        help="Path to MR nifti image", required=True)

    parser.add_argument("-o", metavar="filepath", dest="output",
                        help="Output filepath for segmentation masks",
                        required=True)
    
    parser.add_argument("-d", "--device", choices=["gpu", "cpu", "mps"],
                        help="Device to run on (default: gpu).",
                        default="gpu")
    
    parser.add_argument("-m", metavar="directory", dest="path_to_model",
                        help="Path to trained nnunet model (if not specified looks for " +\
                            "weights at ~/.mrisegmentator_weights or env var $MRISEGMENTATOR_DIR " +\
                            "and will attempt to download if not at either location.",
                        required=False)

    args = parser.parse_args()

    from .inference import mri_segmentator

    mri_segmentator(args.input, args.output, args.path_to_model, args.device)

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

long_description = open('README.md').read()

setup(name='MRISegmentator',
        version='0.4.1',
        description='Robust segmentation of 62 abdominal structures on MRI T1 Weighted images.',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url='https://github.com/rsummers11/MRISegmentator',
        author='Yan Zhuang and Abhinav Suri',
        author_email='rsummers@nih.gov',
        python_requires='>=3.9',
        license='Please See Github',
        packages=find_packages(),
        install_requires=[
            'torch>=2.0.0',
            'tqdm>=4.45.0',
            'p_tqdm',
            'nnunetv2>=2.2.1',
            'requests'
        ],
        zip_safe=False,
        classifiers=[
            'Intended Audience :: Science/Research',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering',
        ],
        entry_points={
            'console_scripts': [
                'MRISegmentator=mrisegmentator.cmdline:main',
                'MRISegmentator_Redownload=mrisegmentator.download_weights:force_redownload',
            ],
        },
    )



"""
function: inference script for the MRISegmentator package
author = "Abhinav Suri, Yan Zhuang"
license = "Please see the license file"
email = "yan.zhuang2@nih.gov or rsummers@mail.cc.nih.gov"
Date: 06/27/2024
"""

from batchgenerators.utilities.file_and_folder_operations import join
import os
import contextlib
import sys
import warnings
from .download_weights import download_weights

class DummyFile:
    def write(self, x): pass
    def flush(self): pass

@contextlib.contextmanager
def nostdout(verbose=False):
    if not verbose:
        save_stdout = sys.stdout
        sys.stdout = DummyFile()
        yield
        sys.stdout = save_stdout
    else:
        yield


def mri_segmentator(input_file, output_file, path_to_model, device='gpu'):

    print("Checking inputs")
    if input_file.find('.nii') == -1:
        raise ValueError("invalid file type: must have .nii in filename")
    if os.path.exists(input_file) == False:
        raise ValueError("path to input file must exist")
    if device not in ['gpu', 'cpu', 'mps']:
        raise ValueError("invalid device: must be one of 'gpu', 'cpu', or 'mps'")

    # nnunet needs absolute path to a file to handle cases where user supplies bare filename

    output_file = os.path.abspath(output_file)

    if path_to_model is None:
        path_to_model = download_weights()

    if not os.path.isdir(path_to_model):
        raise ValueError("invalid path to trained model folder")

    print("setup devices") 

    import torch

    if device == 'gpu':
        devicet = torch.device('cuda')
    elif device == 'cpu':
        devicet = torch.device('cpu')
    elif device == 'mps':
        devicet = torch.device('mps')

    if device == 'cpu':
        import multiprocessing
        torch.set_num_threads(multiprocessing.cpu_count())

    os.environ['nnUNet_raw'] = output_file
    os.environ['nnUNet_preprocessed'] = output_file
    os.environ['nnUNet_results'] = output_file
    # os.environ['nnUNet_preprocessed'] = str(path_to_model)
    # os.environ['nnUNet_results'] = str(path_to_model)

    warnings.filterwarnings("ignore", category=UserWarning, module="nnunetv2")
    with nostdout():
        from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor

    print("Instantiating nnunet predictor...")
    with nostdout():
        predictor = nnUNetPredictor(
            tile_step_size=0.5,
            use_gaussian=True,
            use_mirroring=False,
            perform_everything_on_device=True,
            device=devicet,
            verbose=False,
            verbose_preprocessing=False,
            allow_tqdm=True
        )

   
    print("Instantiating trained model...")
    with nostdout():
        predictor.initialize_from_trained_model_folder(
            path_to_model,
            use_folds=(0,1,2,3,4) if device != 'cpu' else (0,),
            checkpoint_name='checkpoint_final.pth',
        )

    print("Starting prediction...")
    with nostdout():
        predictor.predict_from_files([[input_file]],
             [output_file],
             save_probabilities=False, overwrite=True,
             num_processes_preprocessing=1, num_processes_segmentation_export=1,
             folder_with_segs_from_prev_stage=None, num_parts=1, part_id=0)
    print("Cleaning up...")
    for i in ['dataset.json', 'plans.json', 'predict_from_raw_data_args.json']:
        try:
            os.remove(os.path.dirname(os.path.abspath(output_file)) + '/' + i)
        except Exception as e:
            print(i, e)
            pass
    print("DONE!")
